#include "jserver.h"

thread session_create(jid owner, session_onSend evt, void *evt_data)
{
    /* creats a new session, registers the resource for it, and spawns a thread to handle it */
    thread t;
    session s;
    xmlnode presence;
    jid user;
    pth_attr_t attr;

    if(owner == NULL || owner->resource == NULL) /* illegal */
        return NULL;
 
    jdebug(ZONE,"session_create");
 
    /* create session thread */
    t = thread_new();
    s = pmalloc(t->p, sizeof(struct session_struct));
    t->data = (void *)s;
    s->t = t;

    /* setup internal communication points */
    s->mp_from = pth_msgport_create("session_from");
    s->mp_to = pth_msgport_create("session_to");
    s->send_evt = evt;
    s->send_evt_data = evt_data;

    /* session identity and default settings */
    s->id = jid_new(t->p, jid_full(owner));
    s->u = js_user(owner->user);
    s->r = resource_create(s->u, RESOURCE_SESSION, owner->resource, s->mp_to);
    s->r->aux = (void *)s;
    s->p = t->p;
    s->notification = NULL;
    s->in_count = s->out_count = s->exit_flag = 0;
    s->presence = NULL;

    /* start thread */
    attr = pth_attr_new();
    pth_attr_set(attr, PTH_ATTR_JOINABLE, FALSE);
    pth_attr_set(attr, PTH_ATTR_STACK_SIZE, jlib_stack_default);
    t->id = pth_spawn(attr, session_main, (void *)t);
    pth_attr_destroy(attr);
    pth_yield(NULL); /* let it have a chance to start up */
 
    return t;
}

void session_kill(char *user, char *res, char *reason)
{
    resource r;

    r = resource_lookup(js_user(user), res);
    if(r == NULL || r->type != RESOURCE_SESSION)
        return;

    if(reason != NULL)
        session_end((session)(r->aux), reason);
    else
        session_end((session)(r->aux), "killed");
}

void session_end(session s, char *reason)
{
    xmlnode x;
    jid to_id;

    if(s == NULL || s->exit_flag == 1 || reason == NULL)
        return;

    jdebug(ZONE,"end %d '%s'",s,reason);

    /* flag the session to exit ASAP */
    s->exit_flag = 1;

    /* make sure we're not the primary session */
    s->r->priority = -1;
    js_resetprimary(s->u);

    /* create a presence packet for this action */
    x = xmlnode_new_tag("presence");
    xmlnode_put_attrib(x,"type","unavailable");
    xmlnode_put_attrib(x,"from",jid_full(s->id));
    xmlnode_insert_cdata(xmlnode_insert_tag(x,"status"),reason,strlen(reason));

    /* notify anyone who's been told we're available */
    for(to_id = s->notification; to_id != NULL; to_id = to_id->next)
    {
        xmlnode_put_attrib(x,"to",jid_full(to_id));
        deliver(jpacket_new(xmlnode_dup(x)));
    }

    /* give the session some CPU */
    psend(s->mp_from,jpacket_new(x));
}

void *session_main(void *arg)
{
    thread t = (thread)arg;
    session s = (session)(t->data);
    pth_event_t from, to, ring;
    jpq q;

    jdebug(ZONE,"THREAD:SESSION starting");

    from = pth_event(PTH_EVENT_MSG,s->mp_from);
    to = pth_event(PTH_EVENT_MSG,s->mp_to);
    ring = pth_event_concat(from, to, NULL);

    module_dispatch(mapi_new(M_PHASE_SESSION_ON,t->p,s,NULL));

    while(1)
    {
        pth_wait(ring);

        if(s->exit_flag)
            break;

        while(1)
        {
            /* this loop looks for messages from our user that need to be delivered elsewhere */
            q = (jpq)pth_msgport_get(s->mp_from);
            if(q == NULL)
                break;

            jdebug(ZONE,"THREAD:SESSION:FROM received a message!");
            ++s->out_count;

            /* make sure we have our from set correctly for outgoing crap */
            xmlnode_put_attrib(q->p->x,"from",jid_full(s->id));
            q->p->from = jid_new(q->p->p,jid_full(s->id));

            if(s->u->user != NULL)
                if(session_from_handle(s,q->p) == 1)
                    continue;

            /* pass these to the general delivery function */
            deliver(q->p);
        }

        if(s->exit_flag)
            break;

        while(1)
        {
            /* this loop looks for messages to our user that need to be sent to them */
            q = (jpq)pth_msgport_get(s->mp_to);
            if(q == NULL)
                break;

            jdebug(ZONE,"THREAD:SESSION:TO received data from %s!",jid_full(q->p->from));
            ++s->in_count;

            if(s->u->user != NULL)
                if(session_to_handle(s,q->p) == 1)
                    continue;

            /* we need to check again, s->exit_flag *could* have changed within session_to_handle at some point */
            if(s->exit_flag)
            {
                js_bounce(q->p->x,TERROR_DELIVERY,"Session Disconnected");
                continue;
            }

            /* deliver outgoing for this session to the onSend event passed when this session was created */
            (s->send_evt)(t, q->p);
        }

    }

    jdebug(ZONE,"THREAD:SESSION exiting");

    /* unlist ourselves */
    resource_free(s->r);

    /* make sure the service knows the session is gone */
    (s->send_evt)(t, NULL);

    module_dispatch(mapi_new(M_PHASE_SESSION_OFF,t->p,s,NULL));

    xmlnode_free(s->presence);
    pth_event_free(ring,PTH_FREE_ALL);

    /* the to be processed packets, if any */
    while(1)
    {
        q = (jpq)pth_msgport_get(s->mp_from);
        if(q == NULL)
            break;
        xmlnode_free(q->p->x);
    }
    pth_msgport_destroy(s->mp_from);

    /* the delivery waiting packets, if any */
    while(1)
    {
        q = (jpq)pth_msgport_get(s->mp_to);
        if(q == NULL)
            break;
        if(q->p->type == JPACKET_MESSAGE)
            deliver(q->p); /* re-deliver the messages only */
        else
            xmlnode_free(q->p->x);
    }
    pth_msgport_destroy(s->mp_to);

    logger_session(s);

    thread_free(t);

    return NULL;
}


int session_from_handle(session s, jpacket p)
{
    mapi m;
    xmlnode item, grp;

                switch(p->type)
                {
                    case JPACKET_MESSAGE:
                        m = mapi_new(M_PHASE_MESSAGE_OUT,p->p,s,p);
                        module_dispatch(m);
                        if(m->result != M_DEFAULT)
                        {
                            xmlnode_free(p->x);
                            return 1;
                        }
                        if(j_strcmp(p->to->type,"group") == 0)
                        {
                            for(item = xmlnode_get_firstchild(js_roster(s->u)); item != NULL; item = xmlnode_get_nextsibling(item))
                            {
                                for(grp = xmlnode_get_firstchild(item); grp != NULL; grp = xmlnode_get_nextsibling(grp))
                                    if(j_strcmp(xmlnode_get_name(grp),"group") == 0 && j_strcmp(xmlnode_get_data(xmlnode_get_firstchild(grp)),p->to->server) == 0)
                                        break;
                                if(grp == NULL)
                                    break;
                                xmlnode_put_attrib(p->x,"to",xmlnode_get_attrib(item,"jid"));
                                deliver(jpacket_new(xmlnode_dup(p->x)));
                            }
                            return 1;
                        }
                        break;

                    case JPACKET_PRESENCE:
                        m = mapi_new(M_PHASE_PRESENCE_OUT,p->p,s,p);
                        module_dispatch(m);
                        if(m->result == M_DEFAULT)
                        { /* module didn't handle it, we'll use our internal roster */
                            js_presence_handle(s,p);
                        }
                        return 1;

                    case JPACKET_IQ:
                        /* it was a target send */
                        if(p->to != NULL)
                            return 0;

                        /* hide who we're from since it's internal and only going back to us */
                        p->from = NULL;
                        xmlnode_hide_attrib(p->x,"from");

                        /* known IQ handlers or let module handle it */
                        if(j_strcmp(xmlnode_get_attrib(xmlnode_get_tag(p->x,"query"),"xmlns"),"jabber:iq:roster") == 0)
                            session_roster_handle(s,p);
                        else
                            module_dispatch(mapi_new(M_PHASE_IQ_HANDLE,p->p,s,p));

                        /* reply */
                        js_iqreply(p->x);
                        psend(s->mp_to, p);
                        return 1;
                }

    return 0;
}


int session_to_handle(session s, jpacket p)
{
    mapi m;
    jid to_id;
    xmlnode x;
    int ret = 0;

                switch(p->type)
                {

                    case JPACKET_MESSAGE:
                        m = mapi_new(M_PHASE_MESSAGE_IN,p->p,s,p);
                        module_dispatch(m);
                        if(m->result != M_DEFAULT)
                            ret = 1;
                        break;

                    case JPACKET_PRESENCE:
                        if(jid_cmp(s->id,p->from) == 0)
                        { /* don't forward our own presence */
                            ret = 1;
                            break;
                        }
                        m = mapi_new(M_PHASE_PRESENCE_IN,p->p,s,p);
                        module_dispatch(m);
                        if(m->result == M_DEFAULT)
                        {
                            if(s->presence != NULL && j_strcmp(xmlnode_get_attrib(p->x,"type"),"probe") == 0)
                            {/* answer the probe for the client, if they only have one main presence */
                                x = xmlnode_dup(s->presence);
                                xmlnode_put_attrib(x,"to",jid_full(p->from));
                                xmlnode_put_attrib(x,"from",jid_full(s->id));
                                deliver(jpacket_new(x));
                                ret = 1;
                            }
                        }else{
                            ret = 1;
                        }
                        break;
                    case JPACKET_IQ:
                        /* internal response IQ from mp_from */
                        if(p->to == NULL)
                            break;

                        m = mapi_new(M_PHASE_IQ,p->p,s,p);
                        module_dispatch(m);
                        if(m->result == M_HANDLED)
                        {
                            /* return the result to the sender */
                            js_iqreply(p->x);
                            deliver(jpacket_new(p->x));
                        }
                        if(m->result != M_DEFAULT) /* send to user unless the module took care of it */
                            ret = 1;
                        break;
                }

    if(ret)
        xmlnode_free(p->x);

    return ret;
}

